class InvalidConfigError(Exception):
    pass


class InvalidRaiseSizeError(Exception):
    pass


class InvalidRankError(Exception):
    pass


class InvalidSuitError(Exception):
    pass


class InvalidOrderError(Exception):
    pass


class InvalidHandSizeError(Exception):
    pass


class InvalidHandRankError(Exception):
    pass


class TooFewActivePlayersError(Exception):
    pass


class TableResetError(Exception):
    pass


class InvalidRenderModeError(Exception):
    pass


class RenderInitializationError(Exception):
    pass
